#ifndef DIFFIE_HELMAN_H_INCLUDED
#define DIFFIE_HELMAN_H_INCLUDED

#include "signature.h"
#include "encrypt_decrypt_packet.h"
#include "montgomery.h"
#include "hash_func.h"
#include <stdio.h>

static const u32 dh_length_x32 = 128; // 4096 bit, do not change



//--------------------------------
// No dynamic allocations in this class and nested classes
//
class INITIALIZED_SESSION
{
int initialized;
u32 id[fingerprint_length_x32]; // 256 bit ID of remote side
u32 dh_fingerprint[fingerprint_length_x32]; // 256 bit fingerprint of (g,p,g^x,g^y) for additional verification
ENCRYPT_DECRYPT_PACKET encryptor;
ENCRYPT_DECRYPT_PACKET decryptor;

public:
     INITIALIZED_SESSION();   
    ~INITIALIZED_SESSION();
     void reset();
     u32 *session_id(); // Returns NULL if not initialized, u32[8] if initialized 
     void session_id(u32 *); // Set session ID   
     
     u32 *session_fingerprint(); // Returns NULL if not initialized, u32[8] if initialized
     void session_fingerprint(u32 *); // Set session fingerprint   
     
     void init(u32 const *session_id, u32 const *encrypt_key, u32 const* decrypt_key); // each of keys is u32[18]
     u32 encrypt(u8 *dest, u8 const *src, const u32 src_len_bytes);  // Packet encryption, returns length of encrypted block (= src_length + 16)
     u32 decrypt(u8 *dest, u8 const *src, const u32 src_len_bytes); // length           // Packet decryption, returns length of decrypted block (length -> 0 = invalid block)

     u64 total_rx_bytes();   // total number of bytes supposed to be received (valid + dropped)
     u64 dropped_rx_bytes(); // dropped bytes
     u64 late_packets();     // late packets out of order
     u64 invalid_packets();  // invalid packets (broken, wrong format, invalid MIC)
      
   
};

// Client Side

static const u32 encrypt_decrypt_key_size_x32 = 18; // 576 bits

class TX_SECURE_SESSION
{
u32 len;    
u32 *x; // secret
u32 *p;
u32 *g;
u32 *gxp; // g^x mod p: open data
u32 *gyp; // g^y mod p: open data
u32 *signature;
u32 *far_signature_decoded;
u32 *common_secret;
u8 *tmp_data;

u32 *far_hash;
SERTIFICATE *far_certificate;
SIGNATURE *far_signature;

SIGNATURE *ds;
TEST_GENERATOR_MODULO_P *test_generator_modulo_p;
MONTGOMERY_POW_MOD *mpm;
DIGITAL_SIGNATURE *digital_signature;
COLLAPSE *collapse;
DH_HASH *dh_hash_1;
INITIALIZED_SESSION *initialized_session;

//---- Fixed size ----
int session_is_initialized;
u32 encryption_key[encrypt_decrypt_key_size_x32]; // 576 bit key
u32 decryption_key[encrypt_decrypt_key_size_x32]; // 576 bit key
//--------------------

public:
    TX_SECURE_SESSION(u32 length_x32 = signature_length_x32);
    ~TX_SECURE_SESSION();
    TX_SECURE_SESSION & operator = (const TX_SECURE_SESSION &);
    void reset();
   
    //--------- Sigma-I variant ----------

    u32 init_session_1(u8 *open_data); // Initiate session; returns number of bytes
    u32 process_response_confirm_session_1(u8 *confirmation_data, u8 const *response, // Confirmation data, response 
                         SECRET_DATA const* secret_data, u8 const *password, const u32 password_length_bytes,
                         int (*verify_certificate)(SERTIFICATE const*) = NULL);
    
    u32 get_user_info_1(u8* user_info);
    int session_is_active_1(u32 *session_id);

    //-------------

    int start_initialized_session(INITIALIZED_SESSION *initialized_session); // If DH exchange was successful, we can start initialized session from it 
   
};


// Server side
class RX_SECURE_SESSION
{
u32 len;    
u32 *y; // secret Y
u32 *p;
u32 *g;
u32 *gxp; // g^x mod p: open data
u32 *gyp; // g^x mod p: open data
u32 *signature;
u32 *far_signature_decoded;
u32 *far_hash;

u32 *common_secret;
u8 *tmp_response_data;


DIGITAL_SIGNATURE *digital_signature;
SIGNATURE *ds;
SERTIFICATE *far_certificate;
MONTGOMERY_POW_MOD *mpm;
COLLAPSE *collapse;
DH_HASH *dh_hash_1;
INITIALIZED_SESSION *initialized_session;


//------- fixed size --------
int session_is_initialized;
u32 encryption_key[encrypt_decrypt_key_size_x32];
u32 decryption_key[encrypt_decrypt_key_size_x32];

public:
    RX_SECURE_SESSION(u32 length_x32 = signature_length_x32);
    ~RX_SECURE_SESSION();
    void reset();
    
    //----------- Sigma - I variant ------

    u32 response_to_init_session_1(u8 *response_data, u8 const *open_data, SECRET_DATA const* secret_data,
                                   u8 const *password, const u32 password_length_bytes);

    int check_confirmation_1(u8 const *confirmation, int (*verify_certificate)(SERTIFICATE const*) = NULL);

    u32 get_user_info_1(u8* user_info);

    int session_is_active_1(u32 *session_id = NULL);   
    
    //-----------------------------------

    int start_initialized_session(INITIALIZED_SESSION *initialized_session); // If DH exchange was successful, we can start initialized session from it 

};


//------------------------------------------
// EL GAMAL using certificate as open key
//
class EL_GAMAL_MODE
{
u32 *common_secret;
u32 *open_key;    
u32 *y; // secret key
MONTGOMERY_POW_MOD *mpm;
DIGITAL_SIGNATURE *ds;
COLLAPSE *collapse;
DH_HASH *dh_hash;


u32 encryption_key[encrypt_decrypt_key_size_x32]; // 576 bit key    
u32 data_hash[8]; // 256 bit hash
SIGNATURE signature;

KEYSTREAM_GENERATOR keystream;

void prepare_encryption_data(SERTIFICATE const *sert);
void prepare_decryption_data(SECRET_DATA const* secret_data, u8 const * password, const u32 password_length_bytes);

public:
    EL_GAMAL_MODE();
    ~EL_GAMAL_MODE();

    // Returns number of bytes (0 = error)
    // With optional signing of data

    u64 encrypt(FILE *outfile, FILE *infile, 
                SERTIFICATE const *sert,  // destination  
                SECRET_DATA const *secret_data = NULL,  // signature 
                u8 const *password = NULL, const u32 password_length_bytes = 0);

    
    u64 decrypt(FILE *outfile, FILE *infile, 
                SECRET_DATA const *secret_data, // Recipient key for decryption 
                u8 const *password, const u32 password_length_bytes, // Recipient password
                int (*verify_certificate)(SERTIFICATE const*) = NULL); // Callback for sender verification 

     /* Not defined

     // Returns number of bytes (0 = error)
     // With optional signing of data

    u32 encrypt(u8 *output_data, void const * input_data, const u32 data_length_bytes, 
                                SERTIFICATE const *sert, SECRET_DATA const *secret_data = NULL, 
                                                                       u8 const *password = NULL, const u32 password_length_bytes = 0);
        

    u32 decrypt(u8 *output_data, void const * input_data, const u32 input_data_length_bytes,
                                              SECRET_DATA const *secret_data, // Recipient key for decryption 
                                              u8 const *password, const u32 password_length_bytes, // Recipient password
                                               int (*verify_certificate)(SERTIFICATE const*) = NULL); // Callback for sender verification 
     */
  
    void reset();

};

#endif

