#ifndef NETWORK_H_INCLUDED
#define NETWORK_H_INCLUDED
//-----------------------------------------------------------------------
#include "types.h"
#include "diffie_helman.h"
#include "kademlia.h"
//-----------------------------------------------------------------------
static const u32 NETWORK_SIZE_UDP = 512;
static const u32 NETWORK_SIZE_OBFUSCATION = NETWORK_SIZE_UDP-8;
static const u32 NETWORK_SIZE_SECURED = NETWORK_SIZE_OBFUSCATION-NETWORK_ID_LENGTH*4-4;
static const u32 NETWORK_SIZE_UNSECURED = NETWORK_SIZE_SECURED-16;
static const u32 NETWORK_SIZE_USEFUL = NETWORK_SIZE_UNSECURED-4;
static const u32 NETWORK_SIZE_TEXT = NETWORK_SIZE_USEFUL-32;
static const u32 NETWORK_SIZE_DH_MAX = 8192;
static const u32 NETWORK_PROTOCOL_VERSION = 2;
static const u32 NETWORK_OFFLINE_SEND_LIMIT = 100;
enum {
	NETWORK_PROT_KADEMLIA,
	NETWORK_PROT_CONTROL,
	NETWORK_PROT_FILE_TRANSFER
};
enum {
	USER_STATUS_OFFLINE,
	USER_STATUS_ONLINE,
	USER_STATUS_AWAY,
	USER_STATUS_MAX
};
//-----------------------------------------------------------------------
enum {
	SESSION_ENCODED,
	SESSION_REQUEST,
	SESSION_RESPONSE,
	SESSION_PING_REQ,
	SESSION_PING_RESP,
	SESSION_NAME_REQ,
	SESSION_NAME_RESP,
	SESSION_CONFIRMATION,
	SESSION_TERMINATE,
	SESSION_BUSY,
	SESSION_REJECT,
	SESSION_CONFIRMED,
	SESSION_ENC_FOR,
	SESSION_ENC_FROM,
	SESSION_MAX_PACK
};

#define SESSION_USER_AWAY	0x80000000
//-----------------------------------------------------------------------
enum {
	NET_STATE_IDLE,
	NET_STATE_PINGING,
	NET_STATE_S_REQUESTING,
	NET_STATE_S_CONFIRMING,
	NET_STATE_P_REQUESTED,
	NET_STATE_S_REQUESTED,
	NET_STATE_WAIT_CONFIRMATION,
	NET_STATE_REJECTED,
	NET_STATE_MAX,
};
//-----------------------------------------------------------------------
class ESTABLISHMENT {
public:
	u32 ip;
	u16 port;
	u32 protocol;
	u32 version;
	int state;
	u32 timeout;
	u32 send_requests;
	u32 rec_flags;	// bit flags of received packets
	u32 buff[NETWORK_SIZE_DH_MAX/4];
	u32 last_buf[NETWORK_SIZE_DH_MAX/4];
	u32 sent_buf[NETWORK_SIZE_DH_MAX/4];
	u32 conf_buf[NETWORK_SIZE_DH_MAX/4];
	u32 encapsulated_id[NETWORK_ID_LENGTH];
	u32 rec_size;
	u32 last_size;
	u32 sent_size;
	u32 conf_size;
	TX_SECURE_SESSION tx_session;
	RX_SECURE_SESSION rx_session;
	ESTABLISHMENT();
	~ESTABLISHMENT();
};
//-----------------------------------------------------------------------
class CONNECTION {
public:
	NODE nd;
	bool established;
	u8 userInfo[user_info_max_length_bytes];
	u32 userInfoSize;
	bool away;
	u32 send_cnt;
	u32 pack_num_est;  // number of packets for establishment
	u32 pack_num_sec;  // number of secured packets
	u32 pack_num_drop; // number of dropped packets
	u32 pack_num_sent; // number of sent packets
	u32 encapsulated_id[NETWORK_ID_LENGTH];

	CRITICAL_SECTION est_cs;
	ESTABLISHMENT* est;
	INITIALIZED_SESSION ses;
	CONNECTION* next_connection;
	CONNECTION(const NODE* node);
	~CONNECTION();
};
//-----------------------------------------------------------------------
typedef struct {
	u32* buffer;
	u32 buf_len;
	NODE* nd;
} NET_PACKET;
//-----------------------------------------------------------------------
typedef struct {
	const NODE* nd;
	u32 protocol;
	const u32* buffer;
	u32 buf_len;
	bool dont_count_tx;
} PACKET_FOR_SEND;
//-----------------------------------------------------------------------
class NETWORK {
	u32 sending_buff[NETWORK_SIZE_DH_MAX/4];
	u32 sending_size;
	TX_SECURE_SESSION cmn_tx_session;
	CONNECTION* con;
	HWND mainWnd;
	HANDLE eStopThreads;
	HANDLE recThread;
	HANDLE tmoThread;
	SOCKET soc;
	CRITICAL_SECTION cs;
	u32 my_id[NETWORK_ID_LENGTH];
	u32 enc_host[NETWORK_ID_LENGTH];
	sockaddr_in my_addr;
	bool away;
	bool session_is_initialized;


	friend DWORD WINAPI receiverProc(NETWORK* n);
	friend DWORD WINAPI timeoutProc(NETWORK* n);
	friend DWORD WINAPI rx_signature(CONNECTION* s);
	friend DWORD WINAPI tx_signature(CONNECTION* s);
	friend DWORD WINAPI conf_signature(CONNECTION* s);
	void ProcessRx(u32* buffer_coded,int pack_len,NODE *nd,const u32* host_id);

	CONNECTION* find_connection(const NODE* nd, bool establish);
	CONNECTION* find_connection(const u32* id);
	void send_big_buffer(CONNECTION* con, u32* buf, u32 size, u32 command);
	bool collect_big_buffer(u32* cur_buf, u32 size, ESTABLISHMENT* est);
	void send_small_buffer(CONNECTION *con, bool est, const u32* buf, u32 size);

	void forward_encapsulation(const u32* id_from, const u32* id_to, const u32* buf, u32 size);
	void send_buf(CONNECTION* con, bool est, u32* buf, u32 size);
	void send_to_host(CONNECTION* con, CONNECTION* host, const u32* buf, u32 size);
public:
	NETWORK(NODE* me, HWND mWnd);
	~NETWORK();

	u32 send_packet(PACKET_FOR_SEND* pack);

	CONNECTION* get_first_node(void) {return con;};
	CONNECTION* get_next_node(CONNECTION* cur);
	void get_user_info(NODE* nd, LPTSTR info);
	void get_counters(const NODE* nd, u32* est, u32* sec, u32* drop, u32* sent, u32* missed, u32* late, u32* invalid);
	int get_user_status(const u32* id);
	u16* get_session_hash(const u32* id);
	void get_ip(NODE* nd);
	void set_ip(NODE* nd);
	void reset_sessions(void);
	void set_away(bool status) {away = status;};

	bool findUserByName(u8* userInfo, u16 size, u32* id);
	bool findNode(const u32 *id, NODE *nd);
};
//-----------------------------------------------------------------------
#endif //NETWORK_H_INCLUDED
